package de.titusgames.jglest.playground;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;

import com.jme.bounding.BoundingBox;
import com.jme.image.Image;
import com.jme.image.Texture;
import com.jme.image.Texture2D;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.Node;
import com.jme.scene.Spatial;
import com.jme.scene.TexCoords;
import com.jme.scene.TriMesh;
import com.jme.scene.state.BlendState;
import com.jme.scene.state.MaterialState;
import com.jme.scene.state.TextureState;
import com.jme.system.DisplaySystem;
import com.jme.system.dummy.DummySystemProvider;
import com.jme.util.TextureKey;
import com.jme.util.TextureManager;
import com.jme.util.export.binary.BinaryExporter;
import com.jme.util.geom.BufferUtils;
import com.jme.util.geom.GeometryTool;
import com.jmex.model.animation.KeyframeController;
import com.jmex.model.converters.FormatConverter;

import de.titusgames.jglest.model.G3dModel;
import de.titusgames.jglest.model.GMesh;
import de.titusgames.jglest.xml.parser.G3dXmlReader;

public class GxmlToJme extends FormatConverter
{
   // private static final Logger logger =
   // Logger.getLogger(GxmlToJme.class.getName());

   private Renderer           renderer;
   public static final String PROPERTY_TEXTUREDIR = "texturedir";

   G3dModel                   model               = null;

   /**
    * Converts an G3dXml file to jME format. The syntax is: "G3dXmlToJme file.g3dxml
    * outfile.jme".
    * 
    * @param args
    *           The array of parameters
    */
   public static void main(String[] args)
   {
      DisplaySystem.getDisplaySystem(DummySystemProvider.DUMMY_SYSTEM_IDENTIFIER);
      new GxmlToJme().attemptFileConvert(args);
   }

   /**
    * Converts an .g3dxml file to .jme format. If you wish to use a .mtl to load
    * the g3dxml's material information please specify the base url where the .mtl
    * is located with setProperty("mtllib",new URL(baseURL))
    * 
    * @param format
    *           The .g3dxml file's stream.
    * @param jMEFormat
    *           The .jme file's stream.
    * @throws IOException
    *            If anything bad happens.
    */
   @Override
   public void convert(InputStream format, OutputStream jMEFormat)
         throws IOException
   {
      renderer = DisplaySystem.getDisplaySystem().getRenderer();
      G3dXmlReader greader = new G3dXmlReader();
      try
      {
         model = greader.readModel(format);
      }
      catch (Exception e)
      {
         throw new IOException(e);
      }
      BinaryExporter.getInstance().save(buildStructure(null), jMEFormat);
      nullAll();
   }

   /**
    * Nulls all to let the gc do its job.
    * 
    * @throws IOException
    */
   private void nullAll() throws IOException
   {
      model = null;
      renderer = null;
   }

   /**
    * Converts the structures of the .g3dxml file to a scene to write
    * 
    * @return The TriMesh or Node that represents the .g3dxml file.
    */
   private Spatial buildStructure(String returnedNodeName)
   {
      Node toReturn;
      if (returnedNodeName == null || returnedNodeName == "")
      {
         toReturn = new Node("g3dxml file");
      }
      else
      {
         toReturn = new Node(returnedNodeName);
      }
      int i = 0;
      for (GMesh gmesh : model.getMeshs())
      {
         TriMesh thisMesh = new TriMesh(model.getName() == null
               ? "temp" + i
               : model.getName());
         Vector3f[] vert = gmesh.getVertices().get(0);
         Vector3f[] norm = gmesh.getNormals().get(0);
         Vector2f[] text = gmesh.getTexCoords();
         boolean hasNorm = true, hasTex = true;

         int[] indices = gmesh.getIndices();

         thisMesh.reconstruct(
               BufferUtils.createFloatBuffer(vert), hasNorm
                     ? BufferUtils.createFloatBuffer(norm)
                     : null, null, hasTex ? TexCoords.makeNew(text) : null,
               BufferUtils.createIntBuffer(indices));
         if (getProperty("sillycolors") != null)
            thisMesh.setRandomColors();

         if (gmesh.getFrameCount() > 1)
         {
            KeyframeController vkc = new KeyframeController();
            vkc.setMorphingMesh(thisMesh);
            for (int j = 0; j < gmesh.getFrameCount(); j++)
            {
               TriMesh etm = new TriMesh();
               Vector3f[] vertMorph = gmesh.getVertices().get(j);
               Vector3f[] normMorph = gmesh.getNormals().get(j);

               etm.setVertexBuffer(BufferUtils.createFloatBuffer(vertMorph));
               etm.setNormalBuffer(BufferUtils.createFloatBuffer(normMorph));
               vkc.setKeyframe(j, etm);
            }
            vkc.setActive(true);
            vkc.setSpeed(8.8f);
            thisMesh.addController(vkc);
            // toReturn.addController(vkc);
         }
         else
         {// this can be called here! For animation vertex reducing is not a
            // good idea !!
            GeometryTool.minimizeVerts(thisMesh, GeometryTool.MV_SAME_COLORS
                  | GeometryTool.MV_SAME_NORMALS | GeometryTool.MV_SAME_TEXS);
         }

         // thisMesh.setRenderQueueMode(Renderer.QUEUE_TRANSPARENT);

         // setup the renderer
         MaterialState m;
         BlendState as;
         m = renderer.createMaterialState();

         m.setAmbient(new ColorRGBA(.2f, .2f, .2f, 1));

         Vector3f diffuseColor = gmesh.getDiffuseColor();
         m.setDiffuse(new ColorRGBA(diffuseColor.getX(), diffuseColor.getY(),
               diffuseColor.getZ(), gmesh.getOpacity()));
         m.setAmbient(new ColorRGBA(diffuseColor.getX(), diffuseColor.getY(),
               diffuseColor.getZ(), gmesh.getOpacity()));

         Vector3f specularColor = gmesh.getSpecularColor();
         // is gmesh.getSpecularPower() right for alpha value ????
         m.setSpecular(new ColorRGBA(specularColor.getX(),
               specularColor.getY(), specularColor.getZ(),
               gmesh.getSpecularPower()));

         m.setEnabled(true);

         as = renderer.createBlendState();
         as.setBlendEnabled(true);
         as.setSourceFunction(BlendState.SourceFunction.SourceAlpha);
         as.setDestinationFunction(BlendState.DestinationFunction.OneMinusSourceAlpha);
         as.setTestEnabled(true);
         as.setTestFunction(BlendState.TestFunction.GreaterThan);
         as.setEnabled(true);

         if (as != null)
         {
            thisMesh.setRenderState(as);
            thisMesh.setRenderQueueMode(Renderer.QUEUE_TRANSPARENT);
         }

         // TODO index 0 ???? did I get something wrong in the xml? Is there
         // only on texture per mesh? I think so ....
         if (gmesh.getTexturePaths().get(0) != null)
         {
            TextureState ts = setupTexture(gmesh, 0);
            if (ts != null)
               thisMesh.setRenderState(ts);

         }

         thisMesh.setModelBound(new BoundingBox());
         thisMesh.updateModelBound();

         toReturn.attachChild(thisMesh);
         i++;
      }
      // if (toReturn.getQuantity() == 1)
      // return toReturn.getChild(0);

      return toReturn;

   }

   private TextureState setupTexture(GMesh gmesh, int i)
   {
      try
      {
         TextureState ts;
         URL texdir = (URL) getProperty("texturedir");
         URL texurl = null;
         if (texdir != null)
         {
            //System.out.println("lade von "+texdir+gmesh.getTexturePaths().get(i));
            texurl = new URL(texdir+gmesh.getTexturePaths().get(i));
            //System.out.println("texurl="+texurl.getPath());
         }
         else
         {
            texurl = new File(gmesh.getTexturePaths().get(i)).toURI().toURL();
            //System.out.println("texurl2="+texurl.getPath());
         }
         TextureKey tkey = new TextureKey(texurl, true,
               TextureManager.COMPRESS_BY_DEFAULT
                     ? Image.Format.Guess
                     : Image.Format.GuessNoCompression);
         Texture t = new Texture2D();
         t.setAnisotropicFilterPercent(0.0f);
         t.setMinificationFilter(Texture.MinificationFilter.BilinearNearestMipMap);
         t.setMagnificationFilter(Texture.MagnificationFilter.Bilinear);
         t.setTextureKey(tkey);
         t.setWrap(Texture.WrapMode.Repeat);
         t.setImageLocation(texurl.toString());
         
         

         ts = renderer.createTextureState();
         ts.setTexture(t);
         ts.setEnabled(true);
         return ts;
      }
      catch (MalformedURLException e)
      {
         e.printStackTrace();
      }
      return null;
   }

}